"""
A package with utilities for managing the persistent word analysis across text
versions of a document.  `PersistenceState` is the highest level of the
interface and the part of the system that's most interesting externally.  `Word`s
are also very important.  The current implementation of `Word` only accounts for
how the number of revisions in which a Word is visible.  If persistent word
views (or something similar) is intended to be kept, refactoring will be
necessary.
"""

from .state import State
from .tokens import Tokens, Token
from . import defaults



from nose.tools import eq_

from .. import serializable


def test_type():
    class Foo(serializable.Type):
        def __init__(self, foo, bar):
            self.foo = foo
            self.bar = bar

    foo = Foo(1, "bar")
    eq_(foo, Foo.deserialize(foo))
    eq_(foo, Foo.deserialize(foo.serialize()))


def test_dict():
    d = serializable.Dict()
    d['foo'] = "bar"
    d['derp'] = "herp"

    eq_(d['foo'], "bar")
    assert 'derp' in d

    eq_(d, serializable.Dict.deserialize(d.serialize(), str))



import logging
import time
from itertools import chain

from ...types import Timestamp
from ...util import iteration, none_or
from .collection import Collection

logger = logging.getLogger("mw.database.collections.revisions")


class AllRevisions(Collection):
    def get(self, rev_id, include_page=False):
        """
        Gets a single revisions by ID.  Checks both the ``revision`` and
        ``archive`` tables.  This method throws a :class:`KeyError` if a
        revision cannot be found.

        :Parameters:
            rev_id : int
                Revision ID
            include_page : bool
                Join revision returned against ``page``

        :Returns:
            A revision row
        """
        rev_id = int(rev_id)
        try:
            rev_row = self.db.revisions.get(rev_id, include_page=include_page)
        except KeyError as e:
            rev_row = self.db.archives.get(rev_id)

        return rev_row

    def query(self, *args, **kwargs):
        """
        Queries revisions (excludes revisions to deleted pages)

        :Parameters:
            page_id : int
                Page identifier.  Filter revisions to this page.
            user_id : int
                User identifier.  Filter revisions to those made by this user.
            user_text : str
                User text (user_name or IP address).  Filter revisions to those
                made by this user.
            before : :class:`mw.Timestamp`
                Filter revisions to those made before this timestamp.
            after : :class:`mw.Timestamp`
                Filter revisions to those made after this timestamp.
            before_id : int
                Filter revisions to those with an ID before this ID
            after_id : int
                Filter revisions to those with an ID after this ID
            direction : str
                "newer" or "older"
            limit : int
                Limit the number of results
            include_page : bool
                Join revisions returned against ``page``

        :Returns:
            An iterator over revision rows.
        """

        revisions = self.db.revisions.query(*args, **kwargs)
        archives = self.db.archives.query(*args, **kwargs)

        if 'direction' in kwargs:
            direction = kwargs['direction']
            if direction not in self.DIRECTIONS:
                raise TypeError("direction must be in {0}".format(self.DIRECTIONS))

            if direction == "newer":
                collated_revisions = iteration.sequence(
                    revisions,
                    archives,
                    compare=lambda r1, r2:\
                            (r1['rev_timestamp'], r1['rev_id']) <=
                            (r2['rev_timestamp'], r2['rev_id'])
                )
            else:  # direction == "older"
                collated_revisions = iteration.sequence(
                    revisions,
                    archives,
                    compare=lambda r1, r2:\
                            (r1['rev_timestamp'], r1['rev_id']) >=
                            (r2['rev_timestamp'], r2['rev_id'])
                )
        else:
            collated_revisions = chain(revisions, archives)

        if 'limit' in kwargs:
            limit = kwargs['limit']

            for i, rev in enumerate(collated_revisions):
                yield rev
                if i >= limit:
                    break

        else:
            for rev in collated_revisions:
                yield rev


class Revisions(Collection):
    
    def get(self, rev_id, include_page=False):
        """
        Gets a single revisions by ID.  Checks the ``revision`` table.   This
        method throws a :class:`KeyError` if a revision cannot be found.

        :Parameters:
            rev_id : int
                Revision ID
            include_page : bool
                Join revision returned against ``page``

        :Returns:
            A revision row
        """
        rev_id = int(rev_id)

        query = """
            SELECT *, FALSE AS archived FROM revision
        """
        if include_page:
            query += """
                INNER JOIN page ON page_id = rev_page
            """

        query += " WHERE rev_id = %s"

        cursor.execute(query, [rev_id])

        for row in cursor:
            return row

        raise KeyError(rev_id)

    def query(self, page_id=None, user_id=None, user_text=None,
              before=None, after=None, before_id=None, after_id=None,
              direction=None, limit=None, include_page=False):
        """
        Queries revisions (excludes revisions to deleted pages)

        :Parameters:
            page_id : int
                Page identifier.  Filter revisions to this page.
            user_id : int
                User identifier.  Filter revisions to those made by this user.
            user_text : str
                User text (user_name or IP address).  Filter revisions to those
                made by this user.
            before : :class:`mw.Timestamp`
                Filter revisions to those made before this timestamp.
            after : :class:`mw.Timestamp`
                Filter revisions to those made after this timestamp.
            before_id : int
                Filter revisions to those with an ID before this ID
            after_id : int
                Filter revisions to those with an ID after this ID
            direction : str
                "newer" or "older"
            limit : int
                Limit the number of results
            include_page : bool
                Join revisions returned against ``page``

        :Returns:
            An iterator over revision rows.
        """
        start_time = time.time()

        page_id = none_or(page_id, int)
        user_id = none_or(user_id, int)
        user_text = none_or(user_text, str)
        before = none_or(before, Timestamp)
        after = none_or(after, Timestamp)
        before_id = none_or(before_id, int)
        after_id = none_or(after_id, int)
        direction = none_or(direction, levels=self.DIRECTIONS)
        include_page = bool(include_page)

        query = """
            SELECT *, FALSE AS archived FROM revision
        """

        if include_page:
            query += """
                INNER JOIN page ON page_id = rev_page
            """

        query += """
            WHERE 1
        """
        values = []

        if page_id is not None:
            query += " AND rev_page = %s "
            values.append(page_id)
        if user_id is not None:
            query += " AND rev_user = %s "
            values.append(user_id)
        if user_text is not None:
            query += " AND rev_user_text = %s "
            values.append(user_text)
        if before is not None:
            query += " AND rev_timestamp < %s "
            values.append(before.short_format())
        if after is not None:
            query += " AND rev_timestamp > %s "
            values.append(after.short_format())
        if before_id is not None:
            query += " AND rev_id < %s "
            values.append(before_id)
        if after_id is not None:
            query += " AND rev_id > %s "
            values.append(after_id)

        if direction is not None:
            
            direction = ("ASC " if direction == "newer" else "DESC ")
            
            if before_id != None or after_id != None:
                query += " ORDER BY rev_id {0}, rev_timestamp {0}".format(direction)
            else:
                query += " ORDER BY rev_timestamp {0}, rev_id {0}".format(direction)

        if limit is not None:
            query += " LIMIT %s "
            values.append(limit)

        cursor = self.db.shared_connection.cursor()
        cursor.execute(query, values)
        count = 0
        for row in cursor:
            yield row
            count += 1

        logger.debug("%s revisions read in %s seconds" % (count, time.time() - start_time))


class Archives(Collection):
    def get(self, rev_id):
        """
        Gets a single revisions by ID.  Checks the ``archive`` table. This
        method throws a :class:`KeyError` if a revision cannot be found.

        :Parameters:
            rev_id : int
                Revision ID

        :Returns:
            A revision row
        """
        rev_id = int(rev_id)

        query = """
            SELECT
                ar_id,
                ar_rev_id AS rev_id,
                ar_page_id AS rev_page,
                ar_page_id AS page_id,
                ar_title AS page_title,
                ar_namespace AS page_namespace,
                ar_text_id AS rev_text_id,
                ar_comment AS rev_comment,
                ar_user AS rev_user,
                ar_user_text AS rev_user_text,
                ar_timestamp AS rev_timestamp,
                ar_minor_edit AS rev_minor_edit,
                ar_deleted AS rev_deleted,
                ar_len AS rev_len,
                ar_parent_id AS rev_parent_id,
                ar_sha1 AS rev_sha1,
                TRUE AS archived
            FROM archive
            WHERE ar_rev_id = %s
        """

        cursor.execute(query, [rev_id])
        for row in cursor:
            return row

        raise KeyError(rev_id)

    def query(self, page_id=None, user_id=None, user_text=None,
              before=None, after=None, before_id=None, after_id=None,
              before_ar_id=None, after_ar_id=None,
              direction=None, limit=None, include_page=True):
        """
        Queries archived revisions (revisions of deleted pages)

        :Parameters:
            page_id : int
                Page identifier.  Filter revisions to this page.
            user_id : int
                User identifier.  Filter revisions to those made by this user.
            user_text : str
                User text (user_name or IP address).  Filter revisions to those
                made by this user.
            before : :class:`mw.Timestamp`
                Filter revisions to those made before this timestamp.
            after : :class:`mw.Timestamp`
                Filter revisions to those made after this timestamp.
            before_id : int
                Filter revisions to those with an ID before this ID
            after_id : int
                Filter revisions to those with an ID after this ID
            direction : str
                "newer" or "older"
            limit : int
                Limit the number of results
            include_page : bool
                This field is ignored.  It's only here for compatibility with
                :class:`mw.database.Revision`.

        :Returns:
            An iterator over revision rows.
        """
        page_id = none_or(page_id, int)
        user_id = none_or(user_id, int)
        before = none_or(before, Timestamp)
        after = none_or(after, Timestamp)
        before_id = none_or(before_id, int)
        after_id = none_or(after_id, int)
        direction = none_or(direction, levels=self.DIRECTIONS)
        limit = none_or(limit, int)

        start_time = time.time()
        cursor = self.db.shared_connection.cursor()

        query = """
            SELECT
                ar_id,
                ar_rev_id AS rev_id,
                ar_page_id AS rev_page,
                ar_page_id AS page_id,
                ar_title AS page_title,
                ar_namespace AS page_namespace,
                ar_text_id AS rev_text_id,
                ar_comment AS rev_comment,
                ar_user AS rev_user,
                ar_user_text AS rev_user_text,
                ar_timestamp AS rev_timestamp,
                ar_minor_edit AS rev_minor_edit,
                ar_deleted AS rev_deleted,
                ar_len AS rev_len,
                ar_parent_id AS rev_parent_id,
                ar_sha1 AS rev_sha1,
                TRUE AS archived
            FROM archive
        """

        query += """
            WHERE 1
        """
        values = []

        if page_id is not None:
            query += " AND ar_page_id = %s "
            values.append(page_id)
        if user_id is not None:
            query += " AND ar_user = %s "
            values.append(user_id)
        if user_text is not None:
            query += " AND ar_user_text = %s "
            values.append(user_text)
        if before is not None:
            query += " AND ar_timestamp < %s "
            values.append(before.short_format())
        if after is not None:
            query += " AND ar_timestamp > %s "
            values.append(after.short_format())
        if before_id is not None:
            query += " AND ar_rev_id < %s "
            values.append(before_id)
        if after_id is not None:
            query += " AND ar_rev_id > %s "
            values.append(after_id)
        if before_ar_id is not None:
            query += " AND ar_id < ? "
            values.append(before_ar_id)
        if after_ar_id is not None:
            query += " AND ar_id > ? "
            values.append(after_ar_id)

        if direction is not None:
            dir = ("ASC " if direction == "newer" else "DESC ")
            
            if before is not None or after is not None:
                query += " ORDER BY ar_timestamp {0}, ar_rev_id {0}".format(dir)
            elif before_id is not None or after_id is not None:
                query += " ORDER BY ar_rev_id {0}, ar_timestamp {0}".format(dir)
            else:
                query += " ORDER BY ar_id {0}".format(dir)
        
        if limit is not None:
            query += " LIMIT %s "
            values.append(limit)

        cursor.execute(query, values)
        count = 0
        for row in cursor:
            yield row
            count += 1

        logger.debug("%s revisions read in %s seconds" % (count, time.time() - start_time))



from ...types import serializable, Timestamp
from ...util import none_or

from .comment import Comment
from .contributor import Contributor
from .text import Text
from .util import consume_tags


class Revision(serializable.Type):
    """
    Revision meta data.
    """
    __slots__ = ('id', 'timestamp', 'contributor', 'minor', 'comment', 'text',
                 'bytes', 'sha1', 'parent_id', 'model', 'format')

    TAG_MAP = {
        'id': lambda e: int(e.text),
        'timestamp': lambda e: Timestamp(e.text),
        'contributor': lambda e: Contributor.from_element(e),
        'minor': lambda e: True,
        'comment': lambda e: Comment.from_element(e),
        'text': lambda e: Text.from_element(e),
        'sha1': lambda e: str(e.text),
        'parentid': lambda e: int(e.text),
        'model': lambda e: str(e.text),
        'format': lambda e: str(e.text)
    }

    def __init__(self, id, timestamp, contributor=None, minor=None,
                 comment=None, text=None, bytes=None, sha1=None,
                 parent_id=None, model=None, format=None):
        self.id = none_or(id, int)
        """
        Revision ID : `int`
        """

        self.timestamp = none_or(timestamp, Timestamp)
        """
        Revision timestamp : :class:`mw.Timestamp`
        """

        self.contributor = none_or(contributor, Contributor.deserialize)
        """
        Contributor meta data : :class:`~mw.xml_dump.Contributor` | `None`
        """

        self.minor = False or none_or(minor, bool)
        """
        Is revision a minor change? : `bool`
        """

        self.comment = none_or(comment, str)
        """
        Comment left with revision : :class:`~mw.xml_dump.Comment` (behaves like `str`, with additional members)
        """

        self.text = none_or(text, str)
        """
        Content of text : :class:`~mw.xml_dump.Text` (behaves like `str`, with additional members)
        """

        self.bytes = none_or(bytes, int)
        """
        Number of bytes of content : `str`
        """

        self.sha1 = none_or(sha1, str)
        """
        sha1 hash of the content : `str`
        """

        self.parent_id = none_or(parent_id, int)
        """
        Revision ID of preceding revision : `int` | `None`
        """

        self.model = none_or(model, str)
        """
        TODO: ??? : `str`
        """

        self.format = none_or(format, str)
        """
        TODO: ??? : `str`
        """

    @classmethod
    def from_element(cls, element):
        values = consume_tags(cls.TAG_MAP, element)

        return cls(
            values.get('id'),
            values.get('timestamp'),
            values.get('contributor'),
            values.get('minor') is True,
            values.get('comment'),
            values.get('text'),
            values.get('bytes'),
            values.get('sha1'),
            values.get('parentid'),
            values.get('model'),
            values.get('format')
        )



"""
This module is a collection of utilities for efficiently processing MediaWiki's
XML database dumps.  There are two important concerns that this module intends
to address: *performance* and the *complexity* of streaming XML parsing.

Performance
    Performance is a serious concern when processing large database XML dumps.
    Regretfully, the Global Intepreter Lock prevents us from running threads on
    multiple CPUs.  This library provides a :func:`map`, a function
    that maps a dump processsing over a set of dump files using
    :class:`multiprocessing` to distribute the work over multiple CPUS

Complexity
    Streaming XML parsing is gross.  XML dumps are (1) some site meta data, (2)
    a collection of pages that contain (3) collections of revisions.  The
    module allows you to think about dump files in this way and ignore the
    fact that you're streaming XML.  An :class:`Iterator` contains
    site meta data and an iterator of :class:`Page`'s.  A
    :class:`Page` contains page meta data and an iterator of
    :class:`Revision`'s.  A :class:`Revision` contains revision meta data
    including a :class:`Contributor` (if one a contributor was specified in the
    XML).

"""
from .map import map
from .iteration import *
from .functions import file, open_file



from .detector import Detector
from . import defaults


def detect(checksum_revisions, radius=defaults.RADIUS):
    """
    Detects reverts that occur in a sequence of revisions.  Note that,
    `revision` data meta will simply be returned in the case of a revert.

    This function serves as a convenience wrapper around calls to
    :class:`Detector`'s :meth:`~Detector.process`
    method.

    :Parameters:
        checksum_revisions : iter( ( checksum : str, revision : `mixed` ) )
            an iterable over tuples of checksum and revision meta data
        radius : int
            the maximum revision distance that a revert can span.

    :Return:
        a iterator over :class:`Revert`

    :Example:
        >>> from mw.lib import reverts
        >>>
        >>> checksum_revisions = [
        ...     ("aaa", {'rev_id': 1}),
        ...     ("bbb", {'rev_id': 2}),
        ...     ("aaa", {'rev_id': 3}),
        ...     ("ccc", {'rev_id': 4})
        ... ]
        >>>
        >>> list(reverts.detect(checksum_revisions))
        [Revert(reverting={'rev_id': 3}, reverteds=[{'rev_id': 2}], reverted_to={'rev_id': 1})]

    """

    revert_detector = Detector(radius)

    for checksum, revision in checksum_revisions:
        revert = revert_detector.process(checksum, revision)
        if revert is not None:
            yield revert

# For backwards compatibility
reverts = detect



import os
import re
import subprocess

from .errors import FileTypeError

EXTENSIONS = {
    'xml': ["cat"],
    'gz': ["zcat"],
    'bz2': ["bzcat"],
    '7z': ["7z", "e", "-so"],
    'lzma': ["lzcat"]
}
"""
A map from file extension to the command to run to extract the data to standard out.
"""

EXT_RE = re.compile(r'\.([^\.]+)$')
"""
A regular expression for extracting the final extension of a file.
"""


def file(path_or_f):
    """
    Verifies that a file exists at a given path and that the file has a
    known extension type.

    :Parameters:
        path : `str`
            the path to a dump file

    """
    if hasattr(path_or_f, "readline"):
        return path_or_f
    else:
        path = path_or_f

    path = os.path.expanduser(path)
    if not os.path.isfile(path):
        raise FileTypeError("Can't find file %s" % path)

    match = EXT_RE.search(path)
    if match is None:
        raise FileTypeError("No extension found for %s." % path)
    elif match.groups()[0] not in EXTENSIONS:
        raise FileTypeError("File type %r is not supported." % path)
    else:
        return path


def open_file(path_or_f):
    """
    Turns a path to a dump file into a file-like object of (decompressed)
    XML data.

    :Parameters:
        path : `str`
            the path to the dump file to read
    """
    if hasattr(path_or_f, "read"):
        return path_or_f
    else:
        path = path_or_f

    match = EXT_RE.search(path)
    ext = match.groups()[0]
    p = subprocess.Popen(
        EXTENSIONS[ext] + [path],
        stdout=subprocess.PIPE,
        stderr=open(os.devnull, "w")
    )
    # sys.stderr.write("\n%s %s\n" % (EXTENSIONS[ext], path))
    # sys.stderr.write(p.stdout.read(1000))
    # return False
    return p.stdout



class Dict(dict):
    def __init__(self, *args, vivifier=lambda k: None, **kwargs):
        self.vivifier = vivifier

        dict.__init__(self, *args, **kwargs)

    def __getitem__(self, key):
        if key not in self:
            dict.__setitem__(self, key, self.vivifier(key))

        return dict.__getitem__(self, key)



import xml.etree.ElementTree as etree
from xml.etree.ElementTree import ParseError

from .errors import MalformedXML


def trim_ns(tag):
    return tag[tag.find("}") + 1:]


class EventPointer:
    def __init__(self, etree_events):
        self.tag_stack = []
        self.etree_events = etree_events

    def __next__(self):
        event, element = next(self.etree_events)

        tag = trim_ns(element.tag)

        if event == "start":
            self.tag_stack.append(tag)
        else:
            if self.tag_stack[-1] == tag:
                self.tag_stack.pop()
            else:
                raise MalformedXML("Expected {0}, but saw {1}.".format(
                    self.tag_stack[-1],
                    tag)
                )

        return event, element

    def depth(self):
        return len(self.tag_stack)

    @classmethod
    def from_file(cls, f):
        return EventPointer(etree.iterparse(f, events=("start", "end")))


class ElementIterator:
    def __init__(self, element, pointer):
        self.pointer = pointer
        self.element = element
        self.depth = pointer.depth() - 1

        self.done = False

    def __iter__(self):

        while not self.done and self.pointer.depth() > self.depth:
            event, element = next(self.pointer)

            if event == "start":
                sub_iterator = ElementIterator(element, self.pointer)

                yield sub_iterator

                sub_iterator.clear()

        self.done = True

    def complete(self):

        while not self.done and self.pointer.depth() > self.depth:
            event, element = next(self.pointer)
            if self.pointer.depth() > self.depth:
                element.clear()

        self.done = True

    def clear(self):
        self.complete()
        self.element.clear()

    def attr(self, key, alt=None):
        return self.element.attrib.get(key, alt)

    def __getattr__(self, attr):
        if attr == "tag":
            return trim_ns(self.element.tag)
        elif attr == "text":
            self.complete()
            return self.element.text
        else:
            raise AttributeError("%s has no attribute %r" % (self.__class__.__name__, attr))

    @classmethod
    def from_file(cls, f):
        
        try:
            pointer = EventPointer.from_file(f)
            event, element = next(pointer)
            return cls(element, pointer)
        except ParseError as e:
            raise ParseError(
                    "{0}: {1}...".format(str(e),
                                         str(f.read(500), 'utf-8', 'replace')))



from nose.tools import eq_

from ..functions import reverts


def test_reverts():
    checksum_revisions = [
        ("a", {'id': 1}),
        ("b", {'id': 2}),
        ("c", {'id': 3}),
        ("a", {'id': 4}),
        ("d", {'id': 5}),
        ("b", {'id': 6}),
        ("a", {'id': 7})
    ]

    expected = [
        ({'id': 4}, [{'id': 3}, {'id': 2}], {'id': 1}),
        ({'id': 7}, [{'id': 6}, {'id': 5}], {'id': 4})
    ]

    for revert in reverts(checksum_revisions, radius=2):
        eq_(revert, expected.pop(0))



import time
import logging

import requests

logger = logging.getLogger("mw.util.api.session")

from .functions import none_or


class Session:
    def __init__(self, uri, headers=None, failure_threshold=None, wait_step=2):
        if uri is None:
            raise TypeError("uri must not be None")

        self.uri = str(uri)
        self.headers = headers if headers is not None else {}
        self.session = requests.Session()

        self.failure_threshold = none_or(failure_threshold, int)
        self.wait_step = float(wait_step)

        self.failed = 0

    def __sleep(self):
        time.sleep(self.failed * (self.wait_step ** self.failed))

    def get(self, params, **kwargs):
        return self.request('GET', params, **kwargs)

    def post(self, params, **kwargs):
        return self.request('POST', params, **kwargs)

    def request(self, type, params):
        try:
            result = self.session.request(type, self.uri, params=params)
            self.failed = 0
            return result
        except (requests.HTTPError, requests.ConnectionError, requests.Timeout):
            self.failed += 1

            if self.failed > self.failure_threshold:
                self.failed = 0
                raise
            else:
                self.__sleep()
                self.request(type, params)



"""
This module provides a set of utilities for detecting identity reverts in
revisioned content.

To detect reverts in a stream of revisions to a single page, you can use
:func:`detect`.  If you'll be detecting reverts in a collection of pages or
would, for some other reason, prefer to process revisions one at a time,
:class:`Detector` and it's :meth:`~Detector.process` will allow you to do so.

To detect reverts one-at-time and arbitrarily, you can user the `check()`
functions:

* :func:`database.check` and :func:`database.check_row` use a :class:`mw.database.DB`
* :func:`api.check` and :func:`api.check_rev` use a :class:`mw.api.Session`

Note that these functions are less performant than detecting reverts in a
stream of page revisions.  This can be practical when trying to identify
reverted revisions in a user's contribution history.
"""
from .detector import Detector, Revert
from .functions import detect, reverts
from . import database
from . import api
from . import defaults



from nose.tools import eq_

from ..functions import none_or


def test_none_or():
    eq_(10, none_or("10", int))
    eq_(10.75, none_or("10.75", float))
    eq_(None, none_or(None, int))
    assert none_or("", str) is not None
    assert none_or([], list) is not None
    assert none_or({}, dict) is not None
    assert none_or(0, int) is not None
    assert none_or(-1, int) is not None



def none_or(val, func=None, levels=None):
    if val is None:
        return None
    else:
        if levels is not None:
            if val not in set(levels):
                raise KeyError(val)

            return val
        else:
            return func(val)


def try_keys(dictionary, keys):
    attempted_keys = []
    for key in keys:
        if key in dictionary:
            return dictionary[key]
        attempted_keys.append(key)

    raise KeyError("|".join(str(k) for k in attempted_keys))



from nose.tools import eq_

from ..detector import Detector


def test_detector():
    detector = Detector(2)

    eq_(detector.process("a", {'id': 1}), None)

    # Check noop
    eq_(detector.process("a", {'id': 2}), None)

    # Short revert
    eq_(detector.process("b", {'id': 3}), None)
    eq_(
        detector.process("a", {'id': 4}),
        ({'id': 4}, [{'id': 3}], {'id': 2})
    )

    # Medium revert
    eq_(detector.process("c", {'id': 5}), None)
    eq_(detector.process("d", {'id': 6}), None)
    eq_(
        detector.process("a", {'id': 7}),
        ({'id': 7}, [{'id': 6}, {'id': 5}], {'id': 4})
    )

    # Long (undetected) revert
    eq_(detector.process("e", {'id': 8}), None)
    eq_(detector.process("f", {'id': 9}), None)
    eq_(detector.process("g", {'id': 10}), None)
    eq_(detector.process("a", {'id': 11}), None)



from nose.tools import eq_

from ..text import Text


def test_immutability():
    a = Text("foo")
    b = Text(a)
    eq_(id(a), id(b))


def test_empty_constructor():
    t = Text()
    eq_(t, "")
    eq_(t.deleted, False)
    eq_(t.id, None)
    eq_(t.xml_space, "preserve")
    eq_(t.bytes, None)


def test_deleted_constructor():
    t = Text("", deleted=True)
    eq_(t, "")
    eq_(t.deleted, True)
    eq_(t.id, None)
    eq_(t.xml_space, "preserve")
    eq_(t.bytes, None)


def test_full_constructor():
    t = Text("Foobar!", deleted=False, id=10, xml_space="foobar", bytes=1001)
    eq_(t, "Foobar!")
    eq_(t.deleted, False)
    eq_(t.id, 10)
    eq_(t.xml_space, "foobar")
    eq_(t.bytes, 1001)


def test_serialize():
    t = Text("Foobar!", deleted=False, id=10, xml_space="foobar", bytes=1001)
    t2 = Text.deserialize(t.serialize())
    eq_(t2.deleted, False)
    eq_(t2.id, 10)
    eq_(t2.xml_space, "foobar")
    eq_(t2.bytes, 1001)



from ...types import serializable
from ...util import none_or


class Text(str, serializable.Type):
    """
    Revision text content.  This class behaves identically to
    :class:`str` except that it takes and stores an additional set of parameters.

    **deleted**
        Was the text deleted? : `bool`
    **xml_space**
        What to do with extra whitespace? : `str`
    **id**
        TODO: ??? : `int` | `None`
    **bytes**
        TODO: ??? : `int` | `None`

    >>> from mw.xml_dump import Text
    >>>
    >>> t = Text("foo")
    >>> t == "foo"
    True
    >>> t.deleted
    False
    >>> t.xml_space
    'preserve'
    """

    def __new__(cls, string_or_text="", deleted=False, xml_space="preserve", id=None, bytes=None):
        if isinstance(string_or_text, cls):
            return string_or_text
        else:
            inst = super().__new__(cls, string_or_text)
            inst.initialize(string_or_text, deleted, xml_space, id, bytes)
            return inst

    def initialize(self, string, deleted, xml_space, id, bytes):
        self.deleted = bool(deleted)
        self.xml_space = none_or(xml_space, str)
        self.id = none_or(id, int)
        self.bytes = none_or(bytes, int)

    def __str__(self):
        return str.__str__(self)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join([
                str.__repr__(self),
                "deleted={0}".format(repr(self.deleted))
            ])
        )

    def serialize(self):
        return {
            "string_or_text": str(self),
            "deleted": self.deleted,
            "xml_space": self.xml_space,
            "id": self.id,
            "bytes": self.bytes
        }

    @classmethod
    def from_element(cls, e):
        return cls(
            e.text,
            deleted=e.attr('deleted', False),
            xml_space=e.attr('xml:space'),
            id=e.attr('id'),
            bytes=e.attr('bytes')
        )



from ..util import none_or

from . import serializable


class Namespace(serializable.Type):
    """
    Namespace meta data.
    """

    __slots__ = ('id', 'name', 'aliases', 'case', 'canonical')

    def __init__(self, id, name, canonical=None, aliases=None, case=None):
        self.id = int(id)
        """
        Namespace ID : `int`
        """

        self.name = str(name)
        """
        Namespace name : `str`
        """

        self.aliases = serializable.Set.deserialize(aliases or [], str)
        """
        Alias names : set( `str` )
        """

        self.case = none_or(case, str)
        """
        Case sensitivity : `str` | `None`
        """

        self.canonical = none_or(canonical, str)
        """
        Canonical name : `str` | `None`
        """

    def __hash__(self):
        return self.id



from collections import namedtuple

from ...util import ordered
from . import defaults

Revert = namedtuple("Revert", ['reverting', 'reverteds', 'reverted_to'])
"""
Represents a revert event.  This class behaves like
:class:`collections.namedtuple`.  Note that the datatypes of `reverting`,
`reverteds` and `reverted_to` is not specified since those types will depend
on the revision data provided during revert detection.

:Members:
    **reverting**
        The reverting revision data : `mixed`
    **reverteds**
        The reverted revision data (ordered chronologically) : list( `mixed` )
    **reverted_to**
        The reverted-to revision data : `mixed`
"""


class Detector(ordered.HistoricalMap):
    """
    Detects revert events in a stream of revisions (to the same page) based on
    matching checksums.  To detect reverts, construct an instance of this class and call
    :meth:`process` in chronological order (``direction == "newer"``).

    See `<https://meta.wikimedia.org/wiki/R:Identity_revert>`_

    :Parameters:
        radius : int
            the maximum revision distance that a revert can span.

    :Example:
        >>> from mw.lib import reverts
        >>> detector = reverts.Detector()
        >>>
        >>> detector.process("aaa", {'rev_id': 1})
        >>> detector.process("bbb", {'rev_id': 2})
        >>> detector.process("aaa", {'rev_id': 3})
        Revert(reverting={'rev_id': 3}, reverteds=[{'rev_id': 2}], reverted_to={'rev_id': 1})
        >>> detector.process("ccc", {'rev_id': 4})

    """

    def __init__(self, radius=defaults.RADIUS):
        """
        :Parameters:
            radius : int
                the maximum revision distance that a revert can span.
        """
        super().__init__(maxlen=radius + 1)

    def process(self, checksum, revision=None):
        """
        Process a new revision and detect a revert if it occurred.  Note that
        you can pass whatever you like as `revision` and it will be returned in
        the case that a revert occurs.

        :Parameters:
            checksum : str
                Any identity-machable string-based hash of revision content
            revision : `mixed`
                Revision meta data.  Note that any data will just be returned in the
                case of a revert.

        :Returns:
            a :class:`~mw.lib.reverts.Revert` if one occured or `None`
        """
        revert = None

        if checksum in self:  # potential revert

            reverteds = list(self.up_to(checksum))

            if len(reverteds) > 0:  # If no reverted revisions, this is a noop
                revert = Revert(revision, reverteds, self[checksum])

        self.insert(checksum, revision)
        return revert



from nose.tools import eq_
from ..sequence import sequence


def test_sequence():
    foo = [{'val': 3}, {'val': 5}]
    bar = [{'val': 1}, {'val': 10}, {'val': 15}]
    expected = [{'val': 1}, {'val': 3}, {'val': 5}, {'val': 10}, {'val': 15}]

    result = list(sequence(foo, bar, compare=lambda i1, i2: i1['val'] < i2['val']))
    eq_(expected, result)



def Peekable(it):
    if isinstance(it, PeekableType):
        return it
    else:
        return PeekableType(it)


class PeekableType:
    class EMPTY:
        pass

    def __init__(self, it):
        self.it = iter(it)
        self.__cycle()

    def __iter__(self):
        return self

    def __cycle(self):
        try:
            self.lookahead = next(self.it)
        except StopIteration:
            self.lookahead = self.EMPTY

    def __next__(self):
        item = self.peek()
        self.__cycle()
        return item

    def peek(self):
        if self.empty():
            raise StopIteration()
        else:
            return self.lookahead

    def empty(self):
        return self.lookahead == self.EMPTY



import logging
import time
from itertools import chain

from ...types import Timestamp
from ...util import iteration, none_or
from .collection import Collection

logger = logging.getLogger("mw.database.collections.logging")
'''
class Logging(Collection):

    class Revisions(Collection):

        def query(self,
                  after=None, before=None, after_id=None, befor_id=None
                  page_id=None, user_id=None, user_text=None,
                  namespace=None, title=None, type=None, action=None,
                  direction=None, limit=None):
            """
            Queries revisions (excludes revisions to deleted pages)

            :Parameters:
                page_id : int
                    Page identifier.  Filter revisions to this page.
                user_id : int
                    User identifier.  Filter revisions to those made by this user.
                user_text : str
                    User text (user_name or IP address).  Filter revisions to those
                    made by this user.
                before : :class:`mw.Timestamp`
                    Filter revisions to those made before this timestamp.
                after : :class:`mw.Timestamp`
                    Filter revisions to those made after this timestamp.
                before_id : int
                    Filter revisions to those with an ID before this ID
                after_id : int
                    Filter revisions to those with an ID after this ID
                direction : str
                    "newer" or "older"
                limit : int
                    Limit the number of results

            :Returns:
                An iterator over revision rows.
            """
            start_time = time.time()

            page_id = none_or(page_id, int)
            user_id = none_or(user_id, int)
            user_text = none_or(user_text, str)
            before = none_or(before, Timestamp)
            after = none_or(after, Timestamp)
            before_id = none_or(before_id, int)
            after_id = none_or(after_id, int)
            direction = none_or(direction, levels=self.DIRECTIONS)
            include_page = bool(include_page)

            query = """
                SELECT *, FALSE AS archived FROM revision
            """

            if include_page:
                query += """
                    INNER JOIN page ON page_id = rev_page
                """

            query += """
                WHERE 1
            """
            values = []

            if page_id is not None:
                query += " AND rev_page = ? "
                values.append(page_id)
            if user_id is not None:
                query += " AND rev_user = ? "
                values.append(user_id)
            if user_text is not None:
                query += " AND rev_user_text = ? "
                values.append(user_text)
            if before is not None:
                query += " AND rev_timestamp < ? "
                values.append(before.short_format())
            if after is not None:
                query += " AND rev_timestamp > ? "
                values.append(after.short_format())
            if before_id is not None:
                query += " AND rev_id < ? "
                values.append(before_id)
            if after_id is not None:
                query += " AND rev_id > ? "
                values.append(after_id)

            if direction is not None:
                
                direction = ("ASC " if direction == "newer" else "DESC ")
                
                if before_id != None or after_id != None:
                    query += " ORDER BY rev_id {0}, rev_timestamp {0}".format(direction)
                else:
                    query += " ORDER BY rev_timestamp {0}, rev_id {0}".format(direction)

            if limit is not None:
                query += " LIMIT ? "
                values.append(limit)

            cursor = self.db.shared_connection.cursor()
            cursor.execute(query, values)
            count = 0
            for row in cursor:
                yield row
                count += 1

            logger.debug("%s revisions read in %s seconds" % (count, time.time() - start_time))
'''



import getpass
import logging
import os

import pymysql
import pymysql.cursors

from .collections import AllRevisions, Archives, Pages, Revisions, Users

logger = logging.getLogger("mw.database.db")


class DB:
    """
    Represents a connection to a MySQL database.

    :Parameters:
        connection = :class:`oursql.Connection`
            A connection to a MediaWiki database
    """

    def __init__(self, connection):
        self.shared_connection = connection
        self.shared_connection.cursorclass = pymysql.cursors.DictCursor

        self.revisions = Revisions(self)
        """
        An instance of :class:`mw.database.Revisions`.
        """

        self.archives = Archives(self)
        """
        An instance of :class:`mw.database.Archives`.
        """

        self.all_revisions = AllRevisions(self)
        """
        An instance of :class:`mw.database.AllRevisions`.
        """

        self.pages = Pages(self)
        """
        An instance of :class:`mw.database.Pages`.
        """

        self.users = Users(self)
        """
        An instance of :class:`mw.database.Users`.
        """

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(
                [repr(arg) for arg in self.args] +
                ["%s=%r" % (k, v) for k, v in self.kwargs.items()]
            )
        )

    def __str__(self):
        return self.__repr__()

    @classmethod
    def add_arguments(cls, parser, defaults=None):
        """
        Adds the arguments to an :class:`argparse.ArgumentParser` in order to
        create a database connection.
        """
        defaults = defaults if defaults is not None else defaults

        default_host = defaults.get('host', "localhost")
        parser.add_argument(
            '--host', '-h',
            help="MySQL database host to connect to (defaults to {0})".format(default_host),
            default=default_host
        )

        default_database = defaults.get('database', getpass.getuser())
        parser.add_argument(
            '--database', '-d',
            help="MySQL database name to connect to (defaults to  {0})".format(default_database),
            default=default_database
        )

        default_defaults_file = defaults.get('defaults-file', os.path.expanduser("~/.my.cnf"))
        parser.add_argument(
            '--defaults-file',
            help="MySQL defaults file (defaults to {0})".format(default_defaults_file),
            default=default_defaults_file
        )

        default_user = defaults.get('user', getpass.getuser())
        parser.add_argument(
            '--user', '-u',
            help="MySQL user (defaults to %s)".format(default_user),
            default=default_user
        )
        return parser

    @classmethod
    def from_arguments(cls, args):
        """
        Constructs a :class:`~mw.database.DB`.
        Consumes :class:`argparse.ArgumentParser` arguments given by
        :meth:`add_arguments` in order to create a :class:`DB`.

        :Parameters:
            args : :class:`argparse.Namespace`
                A collection of argument values returned by :class:`argparse.ArgumentParser`'s :meth:`parse_args()`
        """
        connection = pymysql.connect(
            args.host,
            args.user,
            database=args.database,
            read_default_file=args.defaults_file
        )
        return cls(connection)

    @classmethod
    def from_params(cls, *args, **kwargs):
        """
        Constructs a :class:`~mw.database.DB`.  Passes `*args` and `**kwargs`
        to :meth:`oursql.connect` and configures the connection.

        :Parameters:
            args : :class:`argparse.Namespace`
                A collection of argument values returned by :class:`argparse.ArgumentParser`'s :meth:`parse_args()`
        """
        kwargs['cursorclass'] = pymysql.cursors.DictCursor
        if kwargs['db']:
            kwargs['database'] = kwargs['db']
            del kwargs['db']
        connection = pymysql.connect(*args, **kwargs)
        return cls(connection)



import calendar
import datetime
import time

from . import serializable

LONG_MW_TIME_STRING = '%Y-%m-%dT%H:%M:%SZ'
"""
The longhand version of MediaWiki time strings.
"""

SHORT_MW_TIME_STRING = '%Y%m%d%H%M%S'
"""
The shorthand version of MediaWiki time strings.
"""


class Timestamp(serializable.Type):
    """
    An immutable type for working with MediaWiki timestamps in their various
    forms.

    :Parameters:
        time_thing : :class:`mw.Timestamp` | :py:class:`~time.time_struct` | :py:class:`~datetime.datetime` | :py:class:`str` | :py:class:`int` | :py:class:`float`
            The timestamp type from which to construct the timestamp class.

    :Returns:
        :class:`mw.Timestamp`

    You can make use of a lot of different *time things* to initialize a
    :class:`mw.Timestamp`.

    * If a :py:class:`~time.time_struct` or :py:class:`~datetime.datetime` are provided, a `Timestamp` will be constructed from their values.
    * If an `int` or `float` are provided, they will be assumed to a unix timestamp in seconds since Jan. 1st, 1970 UTC.
    * If a `str` is provided, it will be be checked against known MediaWiki timestamp formats.  E.g., ``'%Y%m%d%H%M%S'`` and ``'%Y-%m-%dT%H:%M:%SZ'``.
    * If a :class:`mw.Timestamp` is provided, the same `Timestamp` will be returned.

    For example::

        >>> import datetime, time
        >>> from mw import Timestamp
        >>> Timestamp(1234567890)
        Timestamp('2009-02-13T23:31:30Z')
        >>> Timestamp(1234567890) == Timestamp("2009-02-13T23:31:30Z")
        True
        >>> Timestamp(1234567890) == Timestamp("20090213233130")
        True
        >>> Timestamp(1234567890) == Timestamp(datetime.datetime.utcfromtimestamp(1234567890))
        True
        >>> Timestamp(1234567890) == Timestamp(time.strptime("2009-02-13T23:31:30Z", "%Y-%m-%dT%H:%M:%SZ"))
        True
        >>> Timestamp(1234567890) == Timestamp(Timestamp(1234567890))
        True


    You can also do math and comparisons of timestamps.::

        >>> from mw import Timestamp
        >>> t = Timestamp(1234567890)
        >>> t
        Timestamp('2009-02-13T23:31:30Z')
        >>> t2 = t + 10
        >>> t2
        Timestamp('2009-02-13T23:31:40Z')
        >>> t += 1
        >>> t
        Timestamp('2009-02-13T23:31:31Z')
        >>> t2 - t
        9
        >>> t < t2
        True


    """

    def __new__(cls, time_thing):
        if isinstance(time_thing, cls):
            return time_thing
        elif isinstance(time_thing, time.struct_time):
            return cls.from_time_struct(time_thing)
        elif isinstance(time_thing, datetime.datetime):
            return cls.from_datetime(time_thing)
        elif type(time_thing) in (int, float):
            return cls.from_unix(time_thing)
        else:
            return cls.from_string(time_thing)

    def __init__(self, time_thing):
        # Important that this does nothing in order to allow __new__ to behave
        # as expected.  User `initialize()` instead.
        pass

    def initialize(self, time_struct):
        self.__time = time_struct

    def short_format(self):
        """
        Constructs a long, ``'%Y%m%d%H%M%S'`` formatted string common to the
        database. This method is roughly equivalent to calling
        ``strftime('%Y%m%d%H%M%S')``.

        :Parameters:
            format : str
                The string format

        :Returns:
            A formatted string
        """
        return self.strftime(SHORT_MW_TIME_STRING)

    def long_format(self):
        """
        Constructs a long, ``'%Y-%m-%dT%H:%M:%SZ'`` formatted string common to the
        API. This method is roughly equivalent to calling
        ``strftime('%Y-%m-%dT%H:%M:%SZ')``.

        :Parameters:
            format : str
                The string format

        :Returns:
            A formatted string
        """
        return self.strftime(LONG_MW_TIME_STRING)

    def strftime(self, format):
        """
        Constructs a formatted string.
        See `<https://docs.python.org/3/library/time.html#time.strftime>`_ for a
        discussion of formats descriptors.

        :Parameters:
            format : str
                The format description

        :Returns:
            A formatted string
        """
        return time.strftime(format, self.__time)

    @classmethod
    def strptime(cls, string, format):
        """
        Constructs a :class:`mw.Timestamp` from an explicitly formatted string.
        See `<https://docs.python.org/3/library/time.html#time.strftime>`_ for a
        discussion of formats descriptors.

        :Parameters:
            string : str
                A formatted timestamp
            format : str
                The format description

        :Returns:
            :class:`mw.Timestamp`
        """
        return cls.from_time_struct(time.strptime(string, format))

    @classmethod
    def from_time_struct(cls, time_struct):
        """
        Constructs a :class:`mw.Timestamp` from a :class:`time.time_struct`.

        :Parameters:
            time_struct : :class:`time.time_struct`
                A time structure

        :Returns:
            :class:`mw.Timestamp`
        """
        instance = super().__new__(cls)
        instance.initialize(time_struct)
        return instance

    @classmethod
    def from_datetime(cls, dt):
        """
        Constructs a :class:`mw.Timestamp` from a :class:`datetime.datetime`.

        :Parameters:
            dt : :class:`datetime.datetime``
                A datetime.

        :Returns:
            :class:`mw.Timestamp`
        """
        time_struct = dt.timetuple()
        return cls.from_time_struct(time_struct)

    @classmethod
    def from_unix(cls, seconds):
        """
        Constructs a :class:`mw.Timestamp` from a unix timestamp (in seconds
        since Jan. 1st, 1970 UTC).

        :Parameters:
            seconds : int
                A unix timestamp

        :Returns:
            :class:`mw.Timestamp`
        """
        time_struct = datetime.datetime.utcfromtimestamp(seconds).timetuple()
        return cls.from_time_struct(time_struct)

    @classmethod
    def from_string(cls, string):
        """
        Constructs a :class:`mw.Timestamp` from a MediaWiki formatted string.
        This method is provides a convenient way to construct from common
        MediaWiki timestamp formats. E.g., ``%Y%m%d%H%M%S`` and
        ``%Y-%m-%dT%H:%M:%SZ``.

        :Parameters:
            string : str
                A formatted timestamp

        :Returns:
            :class:`mw.Timestamp`
        """
        if type(string) == bytes:
            string = str(string, 'utf8')
        else:
            string = str(string)

        try:
            return cls.strptime(string, SHORT_MW_TIME_STRING)
        except ValueError as e:
            try:
                return cls.strptime(string, LONG_MW_TIME_STRING)
            except ValueError as e:
                raise ValueError(
                    "{0} is not a valid Wikipedia date format".format(
                        repr(string)
                    )
                )

        return cls.from_time_struct(time_struct)

    def __format__(self, format):
        return self.strftime(format)

    def __str__(self):
        return self.short_format()

    def serialize(self):
        return self.unix()

    @classmethod
    def deserialize(cls, time_thing):
        return Timestamp(time_thing)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            repr(self.long_format())
        )

    def __int__(self):
        return self.unix()

    def __float__(self):
        return float(self.unix())

    def unix(self):
        """
        :Returns:
            the number of seconds since Jan. 1st, 1970 UTC.
        """
        return int(calendar.timegm(self.__time))

    def __sub__(self, other):
        if isinstance(other, Timestamp):
            return self.unix() - other.unix()
        else:
            return self + (other * -1)

    def __add__(self, seconds):
        return Timestamp(self.unix() + seconds)

    def __eq__(self, other):
        try:
            return self.__time == other.__time
        except AttributeError:
            return False

    def __lt__(self, other):
        try:
            return self.__time < other.__time
        except AttributeError:
            return NotImplemented

    def __gt__(self, other):
        try:
            return self.__time > other.__time
        except AttributeError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self.__time <= other.__time
        except AttributeError:
            return NotImplemented

    def __ge__(self, other):
        try:
            return self.__time >= other.__time
        except AttributeError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return not self.__time == other.__time
        except AttributeError:
            return NotImplemented
        
    def __getnewargs__(self):
        return (self.__time,)



import logging

from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.user_contribs")


class UserContribs(Collection):
    """
    A collection of revisions indexes by user.
    """

    PROPERTIES = {'ids', 'title', 'timestamp', 'comment', 'parsedcomment',
                  'size', 'sizediff', 'flags', 'patrolled', 'tags'}

    SHOW = {'minor', '!minor', 'patrolled', '!patrolled'}

    MAX_REVISIONS = 50

    def query(self, *args, limit=None, **kwargs):
        """
        Get a user's revisions.
        See `<https://www.mediawiki.org/wiki/API:Usercontribs>`_

        :Parameters:
            limit : int
                The maximum number of contributions to return.
            start : :class:`mw.Timestamp`
                The start timestamp to return from
            end : :class:`mw.Timestamp`
                The end timestamp to return to
            user : set(str)
                The users to retrieve contributions for.  Maximum number of values 50 (500 for bots)
            userprefix : set(str)
                Retrieve contributions for all users whose names begin with this value.
            direction : str
                "newer" or "older"
            namespace : int
                Only list contributions in these namespaces
            properties :
                Include additional pieces of information

                * ids            - Adds the page ID and revision ID
                * title          - Adds the title and namespace ID of the page
                * timestamp      - Adds the timestamp of the edit
                * comment        - Adds the comment of the edit
                * parsedcomment  - Adds the parsed comment of the edit
                * size           - Adds the new size of the edit
                * sizediff       - Adds the size delta of the edit against its parent
                * flags          - Adds flags of the edit
                * patrolled      - Tags patrolled edits
                * tags           - Lists tags for the edit
            show : set(str)
                Show only items that meet thse criteria, e.g. non minor edits only: ucshow=!minor.
                NOTE: If ucshow=patrolled or ucshow=!patrolled is set, revisions older than
                $wgRCMaxAge (2592000) won't be shown

                * minor
                * !minor,
                * patrolled,
                * !patrolled,
                * top,
                * !top,
                * new,
                * !new
            tag : str
                Only list revisions tagged with this tag
            toponly : bool
                DEPRECATED! Only list changes which are the latest revision
        """
        limit = none_or(limit, int)

        revisions_yielded = 0
        done = False
        while not done:

            if limit is None:
                kwargs['limit'] = self.MAX_REVISIONS
            else:
                kwargs['limit'] = min(limit - revisions_yielded, self.MAX_REVISIONS)

            uc_docs, uccontinue = self._query(*args, **kwargs)

            for doc in uc_docs:
                yield doc
                revisions_yielded += 1

                if limit is not None and revisions_yielded >= limit:
                    done = True
                    break

            if uccontinue is None or len(uc_docs) == 0:
                done = True
            else:
                kwargs['uccontinue'] = uccontinue

    def _query(self, user=None, userprefix=None, limit=None, start=None,
               end=None, direction=None, namespace=None, properties=None,
               show=None, tag=None, toponly=None,
               uccontinue=None):

        params = {
            'action': "query",
            'list': "usercontribs"
        }
        params['uclimit'] = none_or(limit, int)
        params['ucstart'] = self._check_timestamp(start)
        params['ucend'] = self._check_timestamp(end)
        if uccontinue is not None:
            params.update(uccontinue)
        params['ucuser'] = self._items(user, type=str)
        params['ucuserprefix'] = self._items(userprefix, type=str)
        params['ucdir'] = self._check_direction(direction)
        params['ucnamespace'] = none_or(namespace, int)
        params['ucprop'] = self._items(properties, levels=self.PROPERTIES)
        params['ucshow'] = self._items(show, levels=self.SHOW)

        doc = self.session.get(params)
        try:
            if 'query-continue' in doc:
                uccontinue = doc['query-continue']['usercontribs']
            else:
                uccontinue = None

            uc_docs = doc['query']['usercontribs']

            return uc_docs, uccontinue

        except KeyError as e:
            raise MalformedResponse(str(e), doc)



import re


def wikitext_split(text):
    """
    Performs the simplest possible split of latin character-based languages
    and wikitext.

    :Parameters:
        text : str
            Text to split.
    """
    return re.findall(
        r"[\w]+|\[\[|\]\]|\{\{|\}\}|\n+| +|&\w+;|'''|''|=+|\{\||\|\}|\|\-|.",
        text
    )



import logging

from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.revisions")


class Revisions(Collection):
    """
    A collection of revisions indexes by title, page_id and user_text.
    Note that revisions of deleted pages are queriable via
    :class:`mw.api.DeletedRevs`.
    """
    
    PROPERTIES = {'ids', 'flags', 'timestamp', 'user', 'userid', 'size',
                  'sha1', 'contentmodel', 'comment', 'parsedcomment',
                  'content', 'tags', 'flagged'}
    
    DIFF_TO = {'prev', 'next', 'cur'}
    
    # This is *not* the right way to do this, but it should work for all queries.
    MAX_REVISIONS = 50
    
    def get(self, rev_id, **kwargs):
        """
        Get a single revision based on it's ID.  Throws a :py:class:`KeyError`
        if the rev_id cannot be found.
        
        :Parameters:
            rev_id : int
                Revision ID
            ``**kwargs``
                Passed to :py:meth:`query`
            
        :Returns:
            A single rev dict
        """
        rev_id = int(rev_id)
        
        revs = list(self.query(revids={rev_id}, **kwargs))
        
        if len(revs) < 1:
            raise KeyError(rev_id)
        else:
            return revs[0]
        
    def query(self, *args, limit=None, **kwargs):
        """
        Get revision information.
        See `<https://www.mediawiki.org/wiki/API:Properties#revisions_.2F_rv>`_
        
        :Parameters:
            properties : set(str)
                Which properties to get for each revision:
                
                * ids            - The ID of the revision
                * flags          - Revision flags (minor)
                * timestamp      - The timestamp of the revision
                * user           - User that made the revision
                * userid         - User id of revision creator
                * size           - Length (bytes) of the revision
                * sha1           - SHA-1 (base 16) of the revision
                * contentmodel   - Content model id
                * comment        - Comment by the user for revision
                * parsedcomment  - Parsed comment by the user for the revision
                * content        - Text of the revision
                * tags           - Tags for the revision
            limit : int
                Limit how many revisions will be returned
                No more than 500 (5000 for bots) allowed
            start_id : int
                From which revision id to start enumeration (enum)
            end_id : int
                Stop revision enumeration on this revid
            start : :class:`mw.Timestamp`
                From which revision timestamp to start enumeration (enum)
            end : :class:`mw.Timestamp`
                Enumerate up to this timestamp
            direction : str
                "newer" or "older"
            user : str
                Only include revisions made by user_text
            excludeuser : bool
                Exclude revisions made by user
            tag : str
                Only list revisions tagged with this tag
            expandtemplates : bool
                Expand templates in revision content (requires "content" propery)
            generatexml : bool
                Generate XML parse tree for revision content (requires "content" propery)
            parse : bool
                Parse revision content (requires "content" propery)
            section : int
                Only retrieve the content of this section number
            token : set(str)
                Which tokens to obtain for each revision
                
                * rollback - See `<https://www.mediawiki.org/wiki/API:Edit_-_Rollback#Token>`_
            rvcontinue : str
                When more results are available, use this to continue
            diffto : int
                Revision ID to diff each revision to. Use "prev", "next" and
                "cur" for the previous, next and current revision respectively
            difftotext : str
                Text to diff each revision to. Only diffs a limited number of
                revisions. Overrides diffto. If section is set, only that
                section will be diffed against this text
            contentformat : str
                Serialization format used for difftotext and expected for output of content
                
                * text/x-wiki
                * text/javascript
                * text/css
                * text/plain
                * application/json
        
        :Returns:
            An iterator of rev dicts returned from the API.
        """
        
        revisions_yielded = 0
        done = False
        while not done:
            if limit == None:
                kwargs['limit'] = self.MAX_REVISIONS
            else:
                kwargs['limit'] = min(limit - revisions_yielded, self.MAX_REVISIONS)
            
            rev_docs, rvcontinue = self._query(*args, **kwargs)
            
            for doc in rev_docs:
                yield doc
                revisions_yielded += 1
                
                if limit != None and revisions_yielded >= limit:
                    done = True
                    break
                
            if rvcontinue != None and len(rev_docs) > 0:
                kwargs['rvcontinue'] = rvcontinue
            else:
                done = True
            
    
    def _query(self, revids=None, titles=None, pageids=None, properties=None,
                     limit=None, start_id=None, end_id=None, start=None,
                     end=None, direction=None, user=None, excludeuser=None,
                     tag=None, expandtemplates=None, generatexml=None,
                     parse=None, section=None, token=None, rvcontinue=None,
                     diffto=None, difftotext=None, contentformat=None):
        
        params = {
            'action': "query",
            'prop': "revisions"
        }
        
        params['revids'] = self._items(revids, type=int)
        params['titles'] = self._items(titles)
        params['pageids'] = self._items(pageids, type=int)
        
        params['rvprop'] = self._items(properties, levels=self.PROPERTIES)
        
        if revids == None: # Can't have a limit unless revids is none
            params['rvlimit'] = none_or(limit, int)
            
        params['rvstartid'] = none_or(start_id, int)
        params['rvendid'] = none_or(end_id, int)
        params['rvstart'] = self._check_timestamp(start)
        params['rvend'] = self._check_timestamp(end)
        
        params['rvdir'] = self._check_direction(direction)
        params['rvuser'] = none_or(user, str)
        params['rvexcludeuser'] = none_or(excludeuser, int)
        params['rvtag'] = none_or(tag, str)
        params['rvexpandtemplates'] = none_or(expandtemplates, bool)
        params['rvgeneratexml'] = none_or(generatexml, bool)
        params['rvparse'] = none_or(parse, bool)
        params['rvsection'] = none_or(section, int)
        params['rvtoken'] = none_or(token, str)
        params['rvcontinue'] = none_or(rvcontinue, int)
        params['rvdiffto'] = self._check_diffto(diffto)
        params['rvdifftotext'] = none_or(difftotext, str)
        params['rvcontentformat'] = none_or(contentformat, str)
        
        doc = self.session.get(params)
        
        try:
            if 'query-continue' in doc:
                rvcontinue = doc['query-continue']['revisions']['rvcontinue']
            else:
                rvcontinue = None
            
            pages = doc['query']['pages'].values()
            rev_docs = []
            
            for page_doc in pages:
                if 'missing' in page_doc: continue
                
                page_rev_docs = page_doc['revisions']
                del page_doc['revisions']
                
                for rev_doc in page_rev_docs:
                    rev_doc['page'] = page_doc
                
                rev_docs.extend(page_rev_docs)
            
            return rev_docs, rvcontinue
            
        except KeyError as e:
            raise MalformedResponse(str(e), doc)
        
    
    def _check_diffto(self, diffto):
        if diffto == None or diffto in self.DIFF_TO:
            return diffto
        else:
            return int(diffto)



import io

from nose.tools import eq_
from ..element_iterator import EventPointer, ElementIterator


TEST_XML = """
<foo>
    <bar>
        <herp>content</herp>
    </bar>
    <derp foo="bar"></derp>
</foo>
"""


def test_pointer():
    pointer = EventPointer.from_file(io.StringIO(TEST_XML))

    eq_(pointer.tag_stack, [])
    eq_(pointer.depth(), 0)

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo"])
    eq_(pointer.depth(), 1)
    eq_(element.tag, "foo")
    eq_(event, "start")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo", "bar"])
    eq_(pointer.depth(), 2)
    eq_(element.tag, "bar")
    eq_(event, "start")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo", "bar", "herp"])
    eq_(pointer.depth(), 3)
    eq_(element.tag, "herp")
    eq_(event, "start")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo", "bar"])
    eq_(pointer.depth(), 2)
    eq_(element.tag, "herp")
    eq_(event, "end")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo"])
    eq_(pointer.depth(), 1)
    eq_(element.tag, "bar")
    eq_(event, "end")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo", "derp"])
    eq_(pointer.depth(), 2)
    eq_(element.tag, "derp")
    eq_(event, "start")

    event, element = next(pointer)
    eq_(pointer.tag_stack, ["foo"])
    eq_(pointer.depth(), 1)
    eq_(element.tag, "derp")
    eq_(event, "end")

    event, element = next(pointer)
    eq_(pointer.tag_stack, [])
    eq_(pointer.depth(), 0)
    eq_(element.tag, "foo")
    eq_(event, "end")

    try:
        event, element = next(pointer)
    except StopIteration:
        return True

    assert False, "Iteration did not stop as expected."


def test_iterator():
    foo_element = ElementIterator.from_file(io.StringIO(TEST_XML))
    foo_iterator = iter(foo_element)

    bar_element = next(foo_iterator)
    bar_iterator = iter(bar_element)
    eq_(bar_element.tag, "bar")

    herp_element = next(bar_iterator)
    eq_(herp_element.tag, "herp")
    eq_(herp_element.text, "content")

    derp_element = next(foo_iterator)
    eq_(derp_element.tag, "derp")
    eq_(derp_element.attr("foo"), "bar")


def test_skipping_iterator():
    foo_element = ElementIterator.from_file(io.StringIO(TEST_XML))
    foo_iterator = iter(foo_element)

    bar_element = next(foo_iterator)

    derp_element = next(foo_iterator)
    eq_(derp_element.tag, "derp")
    eq_(derp_element.attr("foo"), "bar")



from nose.tools import eq_

from .. import tokenization


def test_wikitext_split():
    eq_(
        list(tokenization.wikitext_split("foo bar herp {{derp}}")),
        ["foo", " ", "bar", " ", "herp", " ", "{{", "derp", "}}"]
    )



import logging

from .cache import Cache
from . import defaults

logger = logging.getLogger("mw.lib.sessions.functions")


def cluster(user_events, cutoff=defaults.CUTOFF):
    """
    Clusters user sessions from a sequence of user events.  Note that,
    `event` data will simply be returned in the case of a revert.

    This function serves as a convenience wrapper around calls to
    :class:`~mw.lib.sessions.Cache`'s :meth:`~mw.lib.sessions.Cache.process`
    method.

    :Parameters:
        user_events : iter( (user, timestamp, event) )
            an iterable over tuples of user, timestamp and event data.

            * user : `hashable`
            * timestamp : :class:`mw.Timestamp`
            * event : `mixed`

        cutoff : int
            the maximum time between events within a user session

    :Returns:
        a iterator over :class:`~mw.lib.sessions.Session`

    :Example:
        >>> from mw.lib import sessions
        >>>
        >>> user_events = [
        ...     ("Willy on wheels", 100000, {'rev_id': 1}),
        ...     ("Walter", 100001, {'rev_id': 2}),
        ...     ("Willy on wheels", 100001, {'rev_id': 3}),
        ...     ("Walter", 100035, {'rev_id': 4}),
        ...     ("Willy on wheels", 103602, {'rev_id': 5})
        ... ]
        >>>
        >>> for user, events in sessions.cluster(user_events):
        ...     (user, events)
        ...
        ('Willy on wheels', [{'rev_id': 1}, {'rev_id': 3}])
        ('Walter', [{'rev_id': 2}, {'rev_id': 4}])
        ('Willy on wheels', [{'rev_id': 5}])


    """

    # Construct the session manager
    cache = Cache(cutoff)

    # Apply the events
    for user, timestamp, event in user_events:

        for session in cache.process(user, timestamp, event):
            yield session

    # Yield the left-overs
    for session in cache.get_active_sessions():
        yield session


# For backwards compatibility
sessions = cluster



import logging

from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.users")


class Users(Collection):
    """
    A collection of information about users
    """

    PROPERTIES = {'blockinfo', 'implicitgroups', 'groups', 'registration',
                  'emailable', 'editcount', 'gender'}

    SHOW = {'minor', '!minor', 'patrolled', '!patrolled'}

    MAX_REVISIONS = 50

    def query(self, *args, **kwargs):
        """
        Get a user's metadata.
        See `<https://www.mediawiki.org/wiki/API:Users>`_

        :Parameters:
            users : str
                The usernames of the users to be retrieved.
            
            properties : set(str)
                Include additional pieces of information

                blockinfo      - Tags if the user is blocked, by whom, and
                                 for what reason
                groups         - Lists all the groups the user(s) belongs to
                implicitgroups - Lists all the groups a user is automatically
                                 a member of
                rights         - Lists all the rights the user(s) has
                editcount      - Adds the user's edit count
                registration   - Adds the user's registration timestamp
                emailable      - Tags if the user can and wants to receive
                                 email through [[Special:Emailuser]]
                gender         - Tags the gender of the user. Returns "male",
                                 "female", or "unknown"
        """
        done = False
        while not done:

            us_docs, query_continue = self._query(*args, **kwargs)

            for doc in us_docs:
                yield doc

            if query_continue is None or len(us_docs) == 0:
                done = True
            else:
                kwargs['query_continue'] = query_continue

    def _query(self, users, query_continue=None, properties=None):

        params = {
            'action': "query",
            'list': "users"
        }
        params['ususers'] = self._items(users, type=str)
        params['usprop'] = self._items(properties, levels=self.PROPERTIES)
        if query_continue is not None:
            params.update(query_continue)

        doc = self.session.get(params)
        try:
            if 'query-continue' in doc:
                query_continue = doc['query-continue']['users']
            else:
                query_continue = None

            us_docs = doc['query']['users']

            return us_docs, query_continue

        except KeyError as e:
            raise MalformedResponse(str(e), doc)



"""
*... iteration ...*
"""
from .iterator import Iterator
from .page import Page
from .redirect import Redirect
from .revision import Revision
from .comment import Comment
from .contributor import Contributor
from .text import Text



def consume_tags(tag_map, element):
    value_map = {}
    for sub_element in element:
        tag_name = sub_element.tag

        if tag_name in tag_map:
            value_map[tag_name] = tag_map[tag_name](sub_element)

    return value_map



class Collection:
    DIRECTIONS = {'newer', 'older'}

    def __init__(self, db):
        self.db = db

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, repr(self.db))



from ...types import serializable
from ...util import none_or

from .util import consume_tags


class Contributor(serializable.Type):
    """
    Contributor meta data.
    """
    __slots__ = ('id', 'user_text')

    TAG_MAP = {
        'id': lambda e: int(e.text),
        'username': lambda e: str(e.text),
        'ip': lambda e: str(e.text)
    }

    def __init__(self, id, user_text):
        self.id = none_or(id, int)
        """
        User ID : int | `None` (if not specified in the XML)
        """

        self.user_text = none_or(user_text, str)
        """
        User name or IP address : str | `None` (if not specified in the XML)
        """

    @classmethod
    def from_element(cls, element):
        values = consume_tags(cls.TAG_MAP, element)

        return cls(
            values.get('id'),
            values.get('username', values.get('ip'))
        )



import io

from nose.tools import eq_
from ....types import Timestamp
from ..iterator import Iterator


SAMPLE_XML = """
<mediawiki xmlns="http://www.mediawiki.org/xml/export-0.8/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http
://www.mediawiki.org/xml/export-0.8/ http://www.mediawiki.org/xml/export-0.8.xsd" version="0.8" xml:lang="en">
  <siteinfo>
    <sitename>Wikipedia</sitename>
    <base>http://en.wikipedia.org/wiki/Main_Page</base>
    <generator>MediaWiki 1.22wmf2</generator>
    <case>first-letter</case>
    <namespaces>
      <namespace key="0" case="first-letter" />
      <namespace key="1" case="first-letter">Talk</namespace>
    </namespaces>
  </siteinfo>
  <page>
    <title>Foo</title>
    <ns>0</ns>
    <id>1</id>
    <revision>
      <id>1</id>
      <timestamp>2004-08-09T09:04:08Z</timestamp>
      <contributor>
        <username>Gen0cide</username>
        <id>92182</id>
      </contributor>
      <text xml:space="preserve">Revision 1 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
    <revision>
      <id>2</id>
      <timestamp>2004-08-10T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.109</ip>
      </contributor>
      <text xml:space="preserve">Revision 2 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <comment>Comment 2</comment>
      <format>text/x-wiki</format>
    </revision>
  </page>
  <page>
    <title>Bar</title>
    <ns>1</ns>
    <id>2</id>
    <redirect title="Computer accessibility" />
    <restrictions>edit=sysop:move=sysop</restrictions>
    <revision>
      <id>3</id>
      <timestamp>2004-08-11T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.22</ip>
      </contributor>
      <text xml:space="preserve">Revision 3 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
    <revision>
      <id>4</id>
      <timestamp>2004-08-12T09:04:08Z</timestamp>
      <text xml:space="preserve">Revision 4 text</text>
      <sha1>6ixvq7o1yg75n9g9chqqg94myzq11c5</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
  </page>
</mediawiki>"""


def test_complete():
    f = io.StringIO(SAMPLE_XML)

    dump = Iterator.from_file(f)
    eq_([0, 1], list(ns.id for ns in dump.namespaces))

    page = next(dump)
    eq_(page.title, "Foo")
    eq_(page.namespace, 0)
    eq_(page.id, 1)
    eq_(page.redirect, None)
    eq_(page.restrictions, [])

    revision = next(page)
    eq_(revision.id, 1)
    eq_(revision.timestamp, Timestamp("2004-08-09T09:04:08Z"))
    eq_(revision.contributor.id, 92182)
    eq_(revision.contributor.user_text, "Gen0cide")
    eq_(revision.text, "Revision 1 text")
    eq_(revision.sha1, "g9chqqg94myzq11c56ixvq7o1yg75n9")
    eq_(revision.comment, None)
    eq_(revision.model, "wikitext")
    eq_(revision.format, "text/x-wiki")

    revision = next(page)
    eq_(revision.id, 2)
    eq_(revision.timestamp, Timestamp("2004-08-10T09:04:08Z"))
    eq_(revision.contributor.id, None)
    eq_(revision.contributor.user_text, "222.152.210.109")
    eq_(revision.text, "Revision 2 text")
    eq_(revision.sha1, "g9chqqg94myzq11c56ixvq7o1yg75n9")
    eq_(revision.comment, "Comment 2")
    eq_(revision.model, "wikitext")
    eq_(revision.format, "text/x-wiki")

    page = next(dump)
    eq_(page.title, "Bar")
    eq_(page.namespace, 1)
    eq_(page.id, 2)
    eq_(page.redirect.title, "Computer accessibility")
    eq_(page.restrictions, ["edit=sysop:move=sysop"])

    revision = next(page)
    eq_(revision.id, 3)
    eq_(revision.timestamp, Timestamp("2004-08-11T09:04:08Z"))
    eq_(revision.contributor.id, None)
    eq_(revision.contributor.user_text, "222.152.210.22")
    eq_(revision.text, "Revision 3 text")
    eq_(revision.sha1, "g9chqqg94myzq11c56ixvq7o1yg75n9")
    eq_(revision.comment, None)
    eq_(revision.model, "wikitext")
    eq_(revision.format, "text/x-wiki")

    revision = next(page)
    eq_(revision.id, 4)
    eq_(revision.timestamp, Timestamp("2004-08-12T09:04:08Z"))
    eq_(revision.contributor, None)
    eq_(revision.text, "Revision 4 text")
    eq_(revision.sha1, "6ixvq7o1yg75n9g9chqqg94myzq11c5")
    eq_(revision.comment, None)
    eq_(revision.model, "wikitext")
    eq_(revision.format, "text/x-wiki")


def test_skipping():
    f = io.StringIO(SAMPLE_XML)

    dump = Iterator.from_file(f)

    page = next(dump)
    eq_(page.title, "Foo")
    eq_(page.namespace, 0)
    eq_(page.id, 1)

    page = next(dump)
    eq_(page.title, "Bar")
    eq_(page.namespace, 1)
    eq_(page.id, 2)

    revision = next(page)
    eq_(revision.id, 3)
    eq_(revision.timestamp, Timestamp("2004-08-11T09:04:08Z"))
    eq_(revision.contributor.id, None)
    eq_(revision.contributor.user_text, "222.152.210.22")
    eq_(revision.text, "Revision 3 text")
    eq_(revision.sha1, "g9chqqg94myzq11c56ixvq7o1yg75n9")
    eq_(revision.comment, None)
    eq_(revision.model, "wikitext")
    eq_(revision.format, "text/x-wiki")


def test_serialization():
    f = io.StringIO(SAMPLE_XML)

    dump = Iterator.from_file(f)

    eq_(dump, Iterator.deserialize(dump.serialize()))



from distutils.core import setup

from setuptools import find_packages

setup(
    name='mediawiki-utilities',
    version=open('VERSION').read().strip(),
    author='Aaron Halfaker',
    author_email='aaron.halfaker@gmail.com',
    packages=find_packages(),
    scripts=[],
    url='http://pypi.python.org/pypi/mediawiki-utilities',
    license=open('LICENSE').read(),
    description='A set of utilities for extracting and processing MediaWiki data.',
    long_description=open('README.rst').read(),
    install_requires=[
        "argparse >= 1.1",
        "requests >= 2.0.1",
        "pymysql >= 0.6.2"],
    test_suite='nose.collector',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: General",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering"
    ],
)



import logging
from multiprocessing import cpu_count, Queue, Value
from queue import Empty

from .functions import file
from .processor import Processor

logger = logging.getLogger("mw.dump.map")


def re_raise(error, path):
    raise error


def map(paths, process_dump, handle_error=re_raise,
        threads=cpu_count(), output_buffer=100):
    """
    Maps a function across a set of dump files and returns
    an (order not guaranteed) iterator over the output.

    The `process_dump` function must return an iterable object (such as a
    generator).  If your process_dump function does not need to produce
    output, make it return an empty `iterable` upon completion (like an empty
    list).

    :Parameters:
        paths : iter( str )
            a list of paths to dump files to process
        process_dump : function( dump : :class:`~mw.xml_dump.Iterator`, path : str)
            a function to run on every :class:`~mw.xml_dump.Iterator`
        threads : int
            the number of individual processing threads to spool up
        output_buffer : int
            the maximum number of output values to buffer.

    :Returns:
        An iterator over values yielded by calls to `process_dump()`
    :Example:
        .. code-block:: python

            from mw import xml_dump

            files = ["examples/dump.xml", "examples/dump2.xml"]

            def page_info(dump, path):
                for page in dump:

                    yield page.id, page.namespace, page.title


            for page_id, page_namespace, page_title in xml_dump.map(files, page_info):
                print("\t".join([str(page_id), str(page_namespace), page_title]))
    """
    pathsq = queue_files(paths)
    outputq = Queue(maxsize=output_buffer)
    running = Value('i', 0)
    threads = max(1, min(int(threads), pathsq.qsize()))

    def dec():
        running.value -= 1
    
    processors = []

    for i in range(0, threads):
        running.value += 1
        processor = Processor(
            pathsq,
            outputq,
            process_dump,
            callback=dec
        )
        processor.start()
        processors.append(processor)

    # output while processes are running
    while running.value > 0:
        try:
            error, item = outputq.get(timeout=.25)
        except Empty:
            continue

        if not error:
            yield item
        else:
            error, processor = item
            re_raise(error, processor)

    # finish yielding output buffer
    try:
        for processor in processors:
            processor.join()
        
        while True:
            error, item = outputq.get(block=False)

            if not error:
                yield item
            else:
                error, processor = item
                re_raise(error, path)

    except Empty:
        pass


def queue_files(paths):
    """
    Produces a `multiprocessing.Queue` containing path for each value in
    `paths` to be used by the `Processor`s.

    :Parameters:
        paths : iterable
            the paths to add to the processing queue
    """
    q = Queue()
    for path in paths:
        q.put(file(path))
    return q



from ...types import serializable
from ...util import none_or


class Redirect(serializable.Type):
    """
    Represents a redirect tag.

    **title**
        Full page name that this page is redirected to : `str`
    """

    def __new__(cls, redirect_or_title):
        if isinstance(redirect_or_title, cls):
            return redirect_or_title
        else:
            inst = super().__new__(cls)
            inst.initialize(redirect_or_title)
            return inst

    def __init__(self, *args, **kwargs):
        pass

    def initialize(self, title):
        self.title = none_or(title, str)

    @classmethod
    def from_element(cls, e):
        return cls(e.attr('title'))



class Token:
    """
    Represents a chunk of text and the revisions of a page that it survived.
    """
    __slots__ = ('text', 'revisions')

    def __init__(self, text, revisions=None):
        self.text = text
        """
        The text of the token.
        """

        self.revisions = revisions if revisions is not None else []
        """
        The meta data for the revisions that the token has appeared within.
        """

    def persist(self, revision):
        self.revisions.append(revision)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join([
                "text={0}".format(repr(self.text)),
                "revisions={0}".format(repr(self.revisions))
            ])
        )


class Tokens(list):
    """
    Represents a :class:`list` of :class:`~mw.lib.persistence.Token` with some
    useful helper functions.

    :Example:

        >>> from mw.lib.persistence import Token, Tokens
        >>>
        >>> tokens = Tokens()
        >>> tokens.append(Token("foo"))
        >>> tokens.extend([Token(" "), Token("bar")])
        >>>
        >>> tokens[0]
        Token(text='foo', revisions=[])
        >>>
        >>> "".join(tokens.texts())
        'foo bar'
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def persist(self, revision):
        for token in self:
            token.persist(revision)

    def texts(self):
        for token in self:
            yield token.text

    def compare(self, new, diff):
        old = self.texts()

        return self.apply_diff(diff(old, new), self, new)

    @classmethod
    def apply_diff(cls, ops, old, new):

        tokens = cls()
        tokens_added = cls()
        tokens_removed = cls()

        for code, a_start, a_end, b_start, b_end in ops:
            if code == "insert":
                for token_text in new[b_start:b_end]:
                    token = Token(token_text)
                    tokens.append(token)
                    tokens_added.append(token)

            elif code == "replace":
                for token_text in new[b_start:b_end]:
                    token = Token(token_text)
                    tokens.append(token)
                    tokens_added.append(token)

                tokens_removed.extend(t for t in old[a_start:a_end])

            elif code == "equal":
                tokens.extend(old[a_start:a_end])
            elif code == "delete":
                tokens_removed.extend(old[a_start:a_end])

            else:
                assert False, \
                    "encounted an unrecognized operation code: " + repr(code)

        return (tokens, tokens_added, tokens_removed)



from .deleted_revs import DeletedRevs
from .pages import Pages
from .recent_changes import RecentChanges
from .revisions import Revisions
from .site_info import SiteInfo
from .user_contribs import UserContribs
from .users import Users



import pickle

from nose.tools import eq_

from ..timestamp import LONG_MW_TIME_STRING, Timestamp


def test_self():
    t1 = Timestamp(1234567890)

    # Unix timestamp
    eq_(t1, Timestamp(int(t1)))

    # Short format
    eq_(t1, Timestamp(t1.short_format()))

    # Long format
    eq_(t1, Timestamp(t1.long_format()))


def test_comparison():
    t1 = Timestamp(1234567890)
    t2 = Timestamp(1234567891)

    assert t1 < t2, "Less than comparison failed"
    assert t2 > t1, "Greater than comparison failed"
    assert not t2 < t1, "Not less than comparison failed"
    assert not t1 > t2, "Not greater than comparison failed"

    assert t1 <= t2, "Less than or equal to comparison failed"
    assert t1 <= t1, "Less than or equal to comparison failed"
    assert t2 >= t1, "Greater than or equal to comparison failed"
    assert t2 >= t2, "Greater than or equal to comparison failed"
    assert not t2 <= t1, "Not less than or equal to comparison failed"
    assert not t1 >= t2, "Not greater than or equal to comparison failed"
    


def test_subtraction():
    t1 = Timestamp(1234567890)
    t2 = Timestamp(1234567891)

    eq_(t2 - t1, 1)
    eq_(t1 - t2, -1)
    eq_(t2 - 1, t1)


def test_strptime():
    eq_(
        Timestamp("2009-02-13T23:31:30Z"),
        Timestamp.strptime("2009-02-13T23:31:30Z", LONG_MW_TIME_STRING)
    )

    eq_(
        Timestamp.strptime(
            "expires 03:20, 21 November 2013 (UTC)",
            "expires %H:%M, %d %B %Y (UTC)"
        ),
        Timestamp("2013-11-21T03:20:00Z")
    )


def test_strftime():
    eq_(
        Timestamp("2009-02-13T23:31:30Z").strftime(LONG_MW_TIME_STRING),
        "2009-02-13T23:31:30Z"
    )

    eq_(
        Timestamp("2009-02-13T23:31:30Z").strftime("expires %H:%M, %d %B %Y (UTC)"),
        "expires 23:31, 13 February 2009 (UTC)"
    )


def test_serialization():
    timestamp = Timestamp(1234567890)
    eq_(
        timestamp,
        Timestamp.deserialize(timestamp.serialize())
    )

def test_pickling():
    timestamp = Timestamp(1234567890)
    eq_(
        timestamp,
        pickle.loads(pickle.dumps(timestamp))
    )



from nose.tools import eq_

from ..state import State


def test_state():
    contents_revisions = [
        ("Apples are red.", 0),
        ("Apples are blue.", 1),
        ("Apples are red.", 2),
        ("Apples are tasty and red.", 3),
        ("Apples are tasty and blue.", 4)
    ]

    state = State()

    token_sets = [state.process(c, r) for c, r in contents_revisions]

    for i, (content, revision) in enumerate(contents_revisions):
        eq_("".join(token_sets[i][0].texts()), content)

    eq_(token_sets[0][0][0].text, "Apples")
    eq_(len(token_sets[0][0][0].revisions), 5)
    eq_(token_sets[0][0][4].text, "red")
    eq_(len(token_sets[0][0][4].revisions), 3)
